/*
 * Show lines from a file.
 *
 * Copyright (C) 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/err.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fsclient/client.h>
#include <systypes/env.h>
#include <systypes/fcntl.h>



/* Read a line from the file. */

static int readline(file_t *file, int *lineno, char **start, char **end)
{
  static char buf[256];
  static offset_t current = 0, limit = 0;
  char *newline = NULL;

  do
  {
    /* Obtain file content. */

    if (!limit)
    {
      limit = client_read(file, buf, 256);

      if (!limit)
      {
        *start = NULL;
        *end = NULL;
        return 0;
      }

      current = 0;

      /* Start of line at start of buffer. */

      if (newline != NULL)
      {
        current += 1;
        *start = buf;
        *end = (char *) memchr(buf + current, (int) '\n', limit - current);

        if (*end == NULL)
          *end = buf + limit;

        return 1;
      }
    }

    /* Find newline. */

    newline = (char *) memchr(buf + current, (int) '\n', limit - current);

    if (newline != NULL)
    {
      (*lineno)++;
      current = newline - (char *) buf + 1;

      /* Start of line before end of buffer. */

      if (current < limit)
      {
        *start = newline + 1;
        *end = (char *) memchr(buf + current, (int) '\n', limit - current);

        if (*end == NULL)
          *end = buf + limit;

        return 1;
      }

      /* Otherwise, reset the buffer to read the start of line. */

      else
        limit = 0;
    }

    /* No newline: read more data. */

    else
      limit = 0;
  }
  while (1);
}



int main(int argc, char *argv[])
{
  file_t *file;
  int i, startline, numlines;
  char *start, *end;

  if (argc < 4)
    return 1;

  file = client_open(argv[1], O_RDONLY);

  if (!client_opened(file))
  {
    if (file != NULL)
      printf("Error: %s\n", l4sys_errtostr(file->error));
    else
      printf("Could not open file.\n");

    while (1);
  }

  startline = atoi(argv[2]);
  numlines = atoi(argv[3]);

  i = 1;
  while (i < startline)
  {
    if (!readline(file, &i, &start, &end))
    {
      printf("EOF error at line %d.\n", i);
      return 1;
    }
  }

  while (i < startline + numlines)
  {
    fwrite(start, sizeof(char), end - start, stdout);
    if (!readline(file, &i, &start, &end))
    {
      printf("EOF error at line %d.\n", i);
      return 1;
    }
  }

  fputs("\n\n", stdout);

  client_close(file);
  client_notifier_close(client_notifier_task());

  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
